from typing import Dict, List, Optional

from dependency_injector.containers import Container

from . import base, module


class Service(base.Service, module.Module):
    def __init__(self, container: Container, name: str):
        super().__init__(container, name)

        self._module_containers: Dict[str, Container] = dict()
        self._modules: Dict[str, module.Module] = dict()

        if hasattr(container, 'modules'):
            for name, provider in container.modules.providers.items():
                module_container = container.modules(name)
                self._module_containers[name] = module_container
                self._modules[name] = module_container.instance(module_container, name)

    @property
    def modules(self) -> List[base.Module]:
        return list(self._modules.values())

    def get_module(self, name) -> Optional[base.Module]:
        return self._modules.get(name)

    def sync_startup(self, application=None, *args, **kwargs):
        super().sync_startup(application=application, service=self, *args, **kwargs)

        for m in self.modules:
            m.sync_startup(application=application, service=self, *args, **kwargs)

    def sync_shutdown(self, application=None, *args, **kwargs):
        for m in self.modules:
            m.sync_shutdown(application=application, service=self, *args, **kwargs)

        super().sync_shutdown(application=application, service=self, *args, **kwargs)

    async def startup(self, application=None, *args, **kwargs):
        await super().startup(application=application, service=self, *args, **kwargs)

        for m in self.modules:
            await m.startup(application=application, service=self, *args, **kwargs)

    async def shutdown(self, application=None, *args, **kwargs):
        for m in self.modules:
            await m.shutdown(application=application, service=self, *args, **kwargs)

        await super().shutdown(application=application, service=self, *args, **kwargs)


class SessionService(Service, module.SessionModule):
    def sync_session_begin(self, application=None, service=None, module=None, *args, **kwargs):
        super().sync_session_begin(application=application, service=self, *args, **kwargs)

        for m in self.modules:
            if not isinstance(m, base.SessionResourceBlock):
                continue
            m.sync_session_begin(application=application, service=self, *args, **kwargs)

    def sync_session_finish(self, application=None, service=None, module=None, *args, **kwargs):
        for m in self.modules:
            if not isinstance(m, base.SessionResourceBlock):
                continue
            m.sync_session_finish(application=application, service=self, *args, **kwargs)

        super().sync_session_finish(application=application, service=self, *args, **kwargs)

    async def session_begin(self, application=None, service=None, module=None, *args, **kwargs):
        await super().session_begin(application=application, service=self, *args, **kwargs)

        for m in self.modules:
            if not isinstance(m, base.SessionResourceBlock):
                continue
            await m.session_begin(application=application, service=self, *args, **kwargs)

    async def session_finish(self, application=None, service=None, module=None, *args, **kwargs):
        for m in self.modules:
            if not isinstance(m, base.SessionResourceBlock):
                continue
            await m.session_finish(application=application, service=self, *args, **kwargs)

        await super().session_finish(application=application, service=self, *args, **kwargs)



import inspect
from dataclasses import dataclass, field
from importlib import import_module
from typing import Union, Callable, Any, Optional, Dict

from .base import ModelProtocol


@dataclass
class Relationship:
    key: Union[str, Callable[[ModelProtocol], Any]]
    model: Union[type, str]
    join_model: Optional[Union[type, str]] = None
    single: bool = True
    on_clause: Optional[Any] = None
    _globals: Dict[str, Any] = field(default_factory=lambda: inspect.stack()[2][0].f_globals, repr=False, compare=False)

    def get_key(self, instance: ModelProtocol):
        if callable(self.key):
            return self.key(instance)
        return getattr(instance, self.key)

    def _get_model(self, model_field) -> ModelProtocol:
        if isinstance(model_field, str):
            if '.' in model_field:
                class_path, class_name = model_field.rsplit('.', maxsplit=1)
                try:
                    module = import_module(class_path)
                    return getattr(module, class_name)
                except ImportError:
                    pass

            if model_field in self._globals:
                return self._globals[model_field]
            else:
                return eval(model_field, self._globals)

        return model_field

    def get_model(self) -> ModelProtocol:
        return self._get_model(self.model)

    def get_join_model(self) -> ModelProtocol:
        return self._get_model(self.join_model or self.model)

    def get_on_clause(self, model):
        if callable(self.on_clause):
            return self.on_clause(model)
        elif isinstance(self.on_clause, str):
            return eval(self.on_clause, self._globals)
        return self.on_clause



from typing import Protocol, Optional, Dict, Any, List


class Serializer(Protocol):
    def serialize(self, data: Optional[Any] = None) -> bytes:
        ...


class Deserializer(Protocol):
    def deserialize(self, data: bytes) -> Optional[Any]:
        ...


class ModelSerializer(Protocol):
    serializer: Serializer

    def serialize(self, only_fields: List[str] = None) -> bytes:
        return self.serializer.serialize(self.get_serializable_values(only_fields))

    def get_serializable_values(self, only_fields: List[str] = None) -> Optional[Dict[str, Any]]:
        ...



import hashlib
import os
import re
from asyncio import get_running_loop, iscoroutinefunction, run as run_main
from functools import partial
from importlib import import_module
from typing import Any, Callable, Dict, Iterable, List

import ujson


def import_path(dotted_path: str):
    if ':' in dotted_path:
        module_path, attr_path = dotted_path.rsplit(':', maxsplit=1)
    else:
        module_path, attr_path = dotted_path.rsplit('.', maxsplit=1)

    module = import_module(module_path)

    if ':' in dotted_path:
        result = module
        for attr_name in attr_path.rsplit('.'):
            result = getattr(result, attr_name)
        return result

    return getattr(module, attr_path)


def lazy_import(dotted_path: str, callable: bool = False, as_partial: bool = False):
    if callable:
        def inner(*args, **kwargs):
            if as_partial:
                return partial(import_path(dotted_path), *args, **kwargs)
            return import_path(dotted_path)(*args, **kwargs)
    else:
        def inner():
            return import_path(dotted_path)
    return inner


def iscoroutinefunction_partial(f):
    while isinstance(f, partial):
        f = f.func
    return iscoroutinefunction(f)


def invoke_sync(func, *args, **kwargs):
    try:
        loop = get_running_loop()
    except RuntimeError:
        loop = None

    if iscoroutinefunction_partial(func):
        if loop:
            return loop.create_task(func(*args, **kwargs))
        else:
            return run_main(func(*args, **kwargs))
    else:
        return func(*args, **kwargs)


async def invoke_async(func, *args, **kwargs):
    if iscoroutinefunction_partial(func):
        return await func(*args, **kwargs)
    else:
        loop = get_running_loop()
        func = partial(func, *args, **kwargs)
        return await loop.run_in_executor(None, func)


def snake2kebab(string: str) -> str:
    return re.sub('_', '-', string)


def kebab2snake(string: str) -> str:
    return re.sub('-', '_', string)


def snake2camel(string: str, upper: bool = False) -> str:
    result = []

    for i, word in enumerate(string.split('_')):
        if upper or i > 0:
            word = word.capitalize()
        result.append(word)

    return ''.join(result)


def default_value(type_: type):
    def getter(value: Any, default: Any):
        return type_(value) if value is not None else default
    return getter


def expandvars_dict(settings_dict):
    """Expands all environment variables in a settings dictionary."""
    return dict((key, os.path.expandvars(value)) for key, value in settings_dict.items())


def prepare_enum_keys(enum_class):
    return [snake2kebab(str(e.value).lower()) for e in enum_class]


def dict_values_to_str(data: dict):
    result = dict()

    for k, v in data.items():
        if isinstance(v, str):
            pass
        elif isinstance(v, dict):
            v = dict_values_to_str(v)
        elif isinstance(v, list):
            v = [str(i) for i in v]
        else:
            v = str(v)

        result[str(k)] = v

    return result


def build_dict_from_dotted_keys(
        iterable: Iterable, key_getter: Callable[[Any], Any],
        value_getter: Callable[[Any], Any])\
        -> Dict[str, Any]:
    result = {}

    for obj in iterable:
        keys = key_getter(obj).split('.')
        cur_dict = result

        for key in keys[:-1]:
            if key not in cur_dict:
                cur_dict[key] = {}

            if not isinstance(cur_dict[key], dict):
                cur_dict[key] = {'.': cur_dict[key]}
            cur_dict = cur_dict[key]

        cur_dict[keys[-1]] = value_getter(obj)

    return result


def build_dotted_keys_from_dict(dict_: Dict[str, Any], root_key: str = None) -> Dict[str, Any]:
    def traverse(key_stack: List[str], values: Dict[str, Any]) -> Dict[str, Any]:
        res = {}

        for k, v in values.items():
            if isinstance(v, dict):
                res.update(traverse(key_stack + [k], v))
            elif isinstance(v, list):
                for i in v:
                    i = traverse(key_stack + [k], i)
                    if isinstance(i, dict):
                        res |= i
            else:
                res['.'.join(key_stack + [k])] = v

        return res

    return traverse([root_key] if root_key else [], dict_)


def obj_dict_to_str_dict(data: dict, value_getter: Callable[[Any], Any]):
    def traverse(values: Dict[str, Any]) -> Dict[str, Any]:
        res = {}

        for k, v in values.items():
            if isinstance(v, dict):
                res[k] = traverse(v)
            elif v is None:
                pass
            else:
                res[k] = value_getter(v)

        return res

    return traverse(data)


def calculate_digest(data: Any, size: int = 0) -> str:
    if isinstance(data, dict):
        dumped = ujson.dumps(dict_values_to_str(data))
    else:
        dumped = str(data)

    dumped = dumped.encode('utf-8')
    size = size if 0 < size <= hashlib.blake2b.MAX_DIGEST_SIZE else 4 + int(len(dumped) / 32)
    digest_size = min(hashlib.blake2b.MAX_DIGEST_SIZE, size)
    hash_sum = hashlib.blake2b(dumped, digest_size=digest_size)
    return hash_sum.hexdigest()



from typing import TypeVar, Any, Union, Dict, Tuple, Optional

from pydantic.main import BaseModel as PydanticModel

from .base import ModelProtocol
from .fields import ModelFieldsMixin
from .local_cache import ModelLocalCacheMixin

CreateSchemaType = TypeVar('CreateSchemaType', bound=PydanticModel)
UpdateSchemaType = TypeVar('UpdateSchemaType', bound=PydanticModel)


class Model(ModelProtocol, ModelLocalCacheMixin, ModelFieldsMixin):
    class Meta(ModelFieldsMixin.Meta):
        pass

    @classmethod
    def has_cache(cls):
        return False

    @classmethod
    async def get_or_create(cls, id_value: Any,
                            data: Union[CreateSchemaType, Dict[str, Any]], **kwargs) \
            -> Tuple[Optional['Model'], bool]:
        if instance := await cls.get(id_value, **kwargs):
            return instance, False
        else:
            return await cls.create(data, **kwargs), True



import os

from pydantic import BaseSettings


class SchedulerSettings(BaseSettings):
    TASK_HOST: str = '0.0.0.0'
    TASK_PORT: int = 3000
    TASK_PROCESSES: int = os.cpu_count()
    TASK_WORKERS: int = os.cpu_count()
    TASK_RPC_TIMEOUT_SECONDS: int = 5
    TASK_RPC_NUM_CONNECT_ATTEMPTS: int = 10
    TASK_CREATE_REMOTE: bool = False

    class Config:
        case_sensitive = False
        env_file = '.env'



import os
import sys

from dependency_injector import containers, providers

from dropland.app.base import ContainerBlock, ResourceBlock
from dropland.log import logger, tr
from dropland.util import default_value
from .engine import EngineConfig, SchedulerBackend


class SchedulerBlock(ContainerBlock, ResourceBlock):
    def sync_startup(self, *args, **kwargs):
        if self.initialized or self.container.as_service():
            return
        self.container.get_scheduler().start()
        logger.info(tr('dropland.blocks.scheduler.started'))
        self._initialized = True

    def sync_shutdown(self, *args, **kwargs):
        if not self.initialized or self.container.as_service():
            return
        self.container.get_scheduler().shutdown(wait=True)
        logger.info(tr('dropland.blocks.scheduler.stopped'))
        self._initialized = False

    async def startup(self, *args, **kwargs):
        if self.initialized or self.container.as_service():
            return
        self.container.get_scheduler().start()
        logger.info(tr('dropland.blocks.scheduler.started'))
        self._initialized = True

    async def shutdown(self, *args, **kwargs):
        if not self.initialized or self.container.as_service():
            return
        self.container.get_scheduler().shutdown(wait=True)
        logger.info(tr('dropland.blocks.scheduler.stopped'))
        self._initialized = False

    def get_engine(self, *args, **kwargs):
        if isinstance(self.container, SchedulerContainer) or 'SchedulerContainer' == self.container.parent_name:
            return self.container.create_engine()
        return self.container.create_engine(*args, **kwargs)

    def get_instance(self):
        return self.container.get_scheduler()


class SimpleSchedulerContainer(containers.DeclarativeContainer):
    __self__ = providers.Self()

    # noinspection PyMethodMayBeStatic
    def _as_service(self):
        return 'worker' == sys.argv[0]

    as_service = providers.Factory(_as_service, __self__)
    engine_factory = providers.Singleton(SchedulerBackend, as_service)

    def _create_engine(self, *args, **kwargs):
        return self.engine_factory().create_engine(*args, **kwargs)

    create_engine = providers.Factory(_create_engine, __self__)
    get_scheduler = providers.Factory(create_engine)
    block = providers.Singleton(SchedulerBlock, __self__)

    wiring_config = containers.WiringConfiguration(
        modules=['.application', '.local']
    )


class SchedulerContainer(SimpleSchedulerContainer):
    __self__ = providers.Self()
    config = providers.Configuration()

    def _create_engine(self):
        if isinstance(self.config.engine_config(), EngineConfig):
            engine_config = self.config.engine_config()
        else:
            engine_config = EngineConfig(
                sql_url=self.config.engine_config.sql_url(),
                sql_tablename=self.config.engine_config.sql_tablename(),
                redis_url=self.config.engine_config.redis_url(),
                redis_job_key=self.config.engine_config.redis_job_key(),
                job_coalesce=self.config.engine_config.job_coalesce.as_(bool)(),
                job_max_instances=self.config.engine_config.
                    job_max_instances.as_(default_value(int))(default=1),
                job_misfire_grace_time=self.config.engine_config.
                    job_misfire_grace_time.as_(default_value(int))(default=24 * 3600),
                task_host=self.config.engine_config.task_host(),
                task_port=self.config.engine_config.task_port(),
                task_processes=self.config.engine_config.
                    task_processes.as_(default_value(int))(default=os.cpu_count()),
                task_workers=self.config.engine_config.
                    task_workers.as_(default_value(int))(default=os.cpu_count()),
                task_rpc_timeout_seconds=self.config.engine_config.
                    task_rpc_timeout_seconds.as_(default_value(int))(default=5),
                task_rpc_num_connect_attempts=self.config.engine_config.
                    task_rpc_num_connect_attempts.as_(default_value(int))(default=10),
                create_remote_engine=self.config.engine_config.create_remote_engine.as_(bool)(),
                timezone=self.config.engine_config.timezone(default='UTC')
            )
        return SimpleSchedulerContainer._create_engine(self, self.config.name(), engine_config)

    create_engine = providers.Factory(_create_engine, __self__)
    get_scheduler = providers.Factory(create_engine)
    block = providers.Singleton(SchedulerBlock, __self__)

    wiring_config = containers.WiringConfiguration(
        modules=['.application', '.local']
    )



# Recommended naming convention used by Alembic, as various different database
# providers will autogenerate vastly different names making migrations more
# difficult. See: http://alembic.zzzcomputing.com/en/latest/naming.html
import enum

NAMING_CONVENTION = {
    'ix': 'ix_%(column_0_label)s',
    'uq': 'uq_%(table_name)s_%(column_0_name)s',
    'ck': 'ck_%(table_name)s_%(constraint_name)s',
    'fk': 'fk_%(table_name)s_%(column_0_name)s_%(referred_table_name)s',
    'pk': 'pk_%(table_name)s'
}


class SqlStorageType(str, enum.Enum):
    SQLITE = 'sqlite'
    POSTGRES = 'postgresql'
    MYSQL = 'mysql'



from dataclasses import dataclass
from datetime import datetime
from random import randint

import pytest
from sqlalchemy import Column, DateTime, Integer, String, func, text

from dropland.blocks.redis import USE_REDIS
from dropland.blocks.sql import USE_SQL
from tests.conftest import sql_storage

pytestmark = pytest.mark.skipif(not USE_SQL, reason='For Sql only')


if USE_SQL:
    from dropland.blocks.sql import SqlStorageEngine, SqlaModel
    from dropland.data.models.sql import CachedSqlModel
    from .sql_models_data import sqla_mysql, sqla_pg, sqla_sqlite
    if USE_REDIS:
        from dropland.blocks.redis.model import RedisCacheType, RedisMethodCache

    @dataclass
    class ModelArgs:
        table_name: str
        sql_engine: SqlStorageEngine
        model_class: type

    ids = ['SQLite', 'PostgreSQL', 'MySQL']
    engines = [sqla_sqlite, sqla_pg, sqla_mysql]
    models_args = []

    for name, engine in zip(ids, engines):
        models_args.append(
            ModelArgs(table_name=f'testmodel_{name.lower()}', sql_engine=engine, model_class=SqlaModel)
        )

    if USE_REDIS:
        from .redis_models_data import redis_engine

        for name, engine in zip(ids, engines):
            table_name = f'testmodel_{name.lower()}_redis_simple'

            class CachedSqlaModel(CachedSqlModel, SqlaModel):
                __abstract__ = True

                class Meta(CachedSqlModel.Meta, SqlaModel.Meta):
                    caches = [RedisMethodCache(redis_engine, table_name, RedisCacheType.SIMPLE)]

            models_args.append(
                ModelArgs(table_name=table_name, sql_engine=engine, model_class=CachedSqlaModel)
            )

        ids += [f'{i}+Redis(simple)' for i in ids]
else:
    models_args, ids = [], []


# noinspection PyPep8Naming
@pytest.mark.asyncio
@pytest.fixture(scope='module', params=models_args, ids=ids)
async def TestModel(request):
    model_args: ModelArgs = request.param
    engine = model_args.sql_engine

    class SomeModel(model_args.model_class, sql_engine=engine):
        __tablename__ = model_args.table_name

        id = Column(Integer, primary_key=True)
        data = Column(String(128))
        create_date = Column(DateTime, server_default=func.now())

        # required in order to access columns with server defaults
        # or SQL expression defaults, subsequent to a flush, without
        # triggering an expired load
        __mapper_args__ = {'eager_defaults': True}

    assert SomeModel.get_engine() is engine
    ti = SomeModel.get_table_info()
    assert ti.db_type == engine.db_type
    assert ti.is_async == engine.is_async

    async with engine.raw_engine.begin() as conn:
        await conn.run_sync(engine.metadata.drop_all, tables=[SomeModel.__table__])
        await conn.run_sync(engine.metadata.create_all, tables=[SomeModel.__table__])
        await conn.commit()

    yield SomeModel

    async with engine.raw_engine.begin() as conn:
        await conn.run_sync(engine.metadata.drop_all, tables=[SomeModel.__table__])
        await conn.commit()


@pytest.fixture(scope='module')
def wire_sql_storage():
    sql_storage.wire()
    yield
    sql_storage.unwire()


@pytest.mark.asyncio
@pytest.fixture(scope='function')
async def model_session(event_loop, wire_sql_storage, redis_session):
    block = sql_storage.block()
    await block.startup()
    await block.session_begin(autocommit=False)
    yield sql_storage.manager()
    await block.session_finish()
    await block.shutdown()


_construct_counter = 0
_construct_list_counter = 0
_build_rela_counter = 0
_build_rela_list_counter = 0


@pytest.fixture(scope='function')
def build_calls_counter(TestModel, monkeypatch):
    global _construct_counter, _construct_list_counter, _build_rela_counter, _build_rela_list_counter
    _construct_counter = 0
    _construct_list_counter = 0
    _build_rela_counter = 0
    _build_rela_list_counter = 0

    old_construct = TestModel._construct
    old_construct_list = TestModel._construct_list
    old_build_rela = TestModel._build_rela
    old_build_rela_list = TestModel._build_rela_list

    def construct(*args, **kwargs):
        global _construct_counter
        _construct_counter += 1
        return old_construct(*args, **kwargs)

    def construct_list(*args, **kwargs):
        global _construct_list_counter
        _construct_list_counter += 1
        return old_construct_list(*args, **kwargs)

    async def build_rela(*args, **kwargs):
        global _build_rela_counter
        _build_rela_counter += 1
        return await old_build_rela(*args, **kwargs)

    async def build_rela_list(*args, **kwargs):
        global _build_rela_list_counter
        _build_rela_list_counter += 1
        return await old_build_rela_list(*args, **kwargs)

    monkeypatch.setattr(TestModel, '_construct', construct)
    monkeypatch.setattr(TestModel, '_construct_list', construct_list)
    monkeypatch.setattr(TestModel, '_build_rela', build_rela)
    monkeypatch.setattr(TestModel, '_build_rela_list', build_rela_list)


@pytest.mark.asyncio
async def test_model_fields(TestModel, model_session):
    m = TestModel()
    assert TestModel.get_fields() == {'id', 'create_date', 'data'}
    assert m.get_values() == {'id': None, 'create_date': None, 'data': None}

    m.id = 123
    m.data = 'qwer'
    assert m.get_values() == {'id': 123, 'create_date': None, 'data': 'qwer'}


@pytest.mark.asyncio
async def test_create(TestModel, build_calls_counter, model_session):
    obj = await TestModel.create(dict(data='123'))
    assert obj.id > 0
    assert obj.data == '123'

    assert _build_rela_counter == 1
    assert _build_rela_list_counter == 0


@pytest.mark.asyncio
async def test_get_or_create(TestModel, build_calls_counter, model_session):
    obj1, created = await TestModel.get_or_create(999, dict(id=999, data='123'))
    assert created is True
    assert obj1.id == 999
    assert obj1.data == '123'

    obj2, created = await TestModel.get_or_create(999, dict(id=999, data='123'))
    assert created is False
    assert obj2.id == 999
    assert obj2.data == '123'

    await TestModel.get(999)

    assert _construct_counter == (3 if TestModel.has_cache() else 1)
    assert _build_rela_counter == _construct_counter
    assert _construct_list_counter == 0
    assert _build_rela_list_counter == 0


@pytest.mark.asyncio
async def test_update(TestModel, build_calls_counter, model_session):
    obj = await TestModel.create(dict(data='123'))
    assert await obj.update(dict(data='456'))
    assert obj.id > 0
    assert obj.data == '456'

    assert _build_rela_counter == 1
    assert _build_rela_list_counter == 0


@pytest.mark.asyncio
async def test_update_by_id(TestModel, build_calls_counter, model_session):
    obj1 = await TestModel.create(dict(data='123'))

    assert await TestModel.update_by_id(obj1.id, dict(data='234'))
    assert not await TestModel.update_by_id(-1, dict(data='234'))

    assert _build_rela_counter == 1
    assert _build_rela_list_counter == 0


@pytest.mark.asyncio
async def test_update_by(TestModel, build_calls_counter, model_session):
    obj1 = await TestModel.create(dict(data='1'))
    obj2 = await TestModel.create(dict(data='2'))
    obj3 = await TestModel.create(dict(data='3'))

    res = await TestModel.update_by([TestModel.data != '1'], dict(data='0'))
    assert res == 2

    assert obj1.data == '1'
    assert obj2.data == '0'
    assert obj3.data == '0'

    assert _build_rela_counter == 3
    assert _build_rela_list_counter == 0


@pytest.mark.asyncio
async def test_get(TestModel, build_calls_counter, model_session):
    obj1 = await TestModel.create(dict(data='123'))

    obj2 = await TestModel.get(obj1.id)
    assert obj1.id == obj2.id
    assert obj1.data == obj2.data
    assert obj1 == obj2

    obj3 = await TestModel.get(-1)
    assert obj3 is None

    assert _construct_counter == (2 if TestModel.has_cache() else 1)
    assert _build_rela_counter == _construct_counter
    assert _construct_list_counter == 0
    assert _build_rela_list_counter == 0


@pytest.mark.asyncio
async def test_list(TestModel, build_calls_counter, model_session):
    obj1 = await TestModel.create(dict(data='1'))
    obj2 = await TestModel.create(dict(data='2'))
    obj3 = await TestModel.create(dict(data='3'))

    objects = await TestModel.list()
    assert len(objects) == 3
    assert objects[0] == obj1
    assert objects[1] == obj2
    assert objects[2] == obj3

    objects = await TestModel.list(skip=1)
    assert len(objects) == 2
    assert objects[0] == obj2
    assert objects[1] == obj3

    objects = await TestModel.list(skip=2)
    assert len(objects) == 1
    assert objects[0] == obj3

    objects = await TestModel.list(skip=3)
    assert len(objects) == 0

    objects = await TestModel.list(limit=1)
    assert len(objects) == 1
    assert objects[0] == obj1

    objects = await TestModel.list(limit=2)
    assert len(objects) == 2
    assert objects[0] == obj1
    assert objects[1] == obj2

    objects = await TestModel.list(limit=3)
    assert len(objects) == 3
    assert objects[0] == obj1
    assert objects[1] == obj2
    assert objects[2] == obj3

    objects = await TestModel.list(skip=1, limit=3)
    assert len(objects) == 2
    assert objects[0] == obj2
    assert objects[1] == obj3

    objects = await TestModel.list(skip=2, limit=1)
    assert len(objects) == 1
    assert objects[0] == obj3

    objects = await TestModel.list(skip=3, limit=3)
    assert len(objects) == 0

    objects = await TestModel.list(skip=2, limit=0)
    assert len(objects) == 1
    assert objects[0] == obj3

    assert _construct_counter == 19
    assert _build_rela_counter == 3
    assert _construct_list_counter == 11
    assert _build_rela_list_counter == 9


@pytest.mark.asyncio
async def test_filter(TestModel, build_calls_counter, model_session):
    obj1 = await TestModel.create(dict(data='1'))
    obj2 = await TestModel.create(dict(data='2'))
    obj3 = await TestModel.create(dict(data='3'))

    objects = await TestModel.list()
    assert len(objects) == 3
    assert objects[0] == obj1
    assert objects[1] == obj2
    assert objects[2] == obj3

    objects = await TestModel.list([])
    assert len(objects) == 3
    assert objects[0] == obj1
    assert objects[1] == obj2
    assert objects[2] == obj3

    objects = await TestModel.list([TestModel.data == '-1'])
    assert len(objects) == 0

    objects = await TestModel.list([TestModel.data == '1'])
    assert len(objects) == 1
    assert objects[0] == obj1

    objects = await TestModel.list([TestModel.data == '1', TestModel.data == '2'])
    assert len(objects) == 0

    objects = await TestModel.list([(TestModel.data == '1') | (TestModel.data == '2')])
    assert len(objects) == 2
    assert objects[0] == obj1
    assert objects[1] == obj2

    objects = await TestModel.list([(TestModel.data == '1') & (TestModel.id == obj1.id)])
    assert len(objects) == 1
    assert objects[0] == obj1

    objects = await TestModel.list([text('data > :data')], params=dict(data='1'))
    assert len(objects) == 2
    assert objects[0] == obj2
    assert objects[1] == obj3

    assert _construct_counter == 15
    assert _build_rela_counter == 3
    assert _construct_list_counter == 8
    assert _build_rela_list_counter == 6


@pytest.mark.asyncio
async def test_any(TestModel, build_calls_counter, model_session):
    obj1 = await TestModel.create(dict(data='1'))
    obj2 = await TestModel.create(dict(data='2'))
    obj3 = await TestModel.create(dict(data='3'))

    objects = await TestModel.get_any([])
    assert len(objects) == 0

    objects = await TestModel.get_any([obj1.id, obj2.id, obj3.id])
    assert len(objects) == 3
    assert objects[0] == obj1
    assert objects[1] == obj2
    assert objects[2] == obj3

    objects = await TestModel.get_any([obj1.id, obj3.id])
    assert len(objects) == 2
    assert objects[0] == obj1
    assert objects[1] == obj3

    objects = await TestModel.get_any([-1, obj1.id, 999999, obj3.id, 0])
    assert len(objects) == 5
    assert objects[0] is None
    assert objects[1] == obj1
    assert objects[2] is None
    assert objects[3] == obj3
    assert objects[4] is None

    await TestModel.get_any([obj1.id, obj3.id])

    assert _construct_counter == (12 if TestModel.has_cache() else 3)
    assert _build_rela_counter == 3
    assert _construct_list_counter == (4 if TestModel.has_cache() else 1)
    assert _build_rela_list_counter == (4 if TestModel.has_cache() else 0)


@pytest.mark.asyncio
async def test_count(TestModel, build_calls_counter, model_session):
    obj1 = await TestModel.create(dict(data='1'))
    obj2 = await TestModel.create(dict(data='2'))
    obj3 = await TestModel.create(dict(data='3'))

    res = await TestModel.count()
    assert res == 3

    res = await TestModel.count([])
    assert res == 3

    res = await TestModel.count([TestModel.data == '-1'])
    assert res == 0

    res = await TestModel.count([TestModel.data == '1'])
    assert res == 1

    res = await TestModel.count([TestModel.data == '1', TestModel.data == '2'])
    assert res == 0

    res = await TestModel.count([(TestModel.data == '1') | (TestModel.data == '2')])
    assert res == 2

    res = await TestModel.count([(TestModel.data == '1') & (TestModel.id == obj1.id)])
    assert res == 1

    res = await TestModel.count([text('data > :data')], params=dict(data='1'))
    assert res == 2

    assert _construct_counter == 3
    assert _build_rela_counter == _construct_counter
    assert _construct_list_counter == 0
    assert _build_rela_list_counter == _construct_list_counter


@pytest.mark.asyncio
async def test_exists(TestModel, build_calls_counter, model_session):
    obj1 = await TestModel.create(dict(data='1'))
    obj2 = await TestModel.create(dict(data='2'))
    obj3 = await TestModel.create(dict(data='3'))

    res = await TestModel.exists(obj1.id)
    assert res is True

    res = await TestModel.exists(-1)
    assert res is False

    res = await TestModel.exists_by([])
    assert res is False

    res = await TestModel.exists_by([TestModel.data == '-1'])
    assert res is False

    res = await TestModel.exists_by([TestModel.data == '1'])
    assert res is True

    res = await TestModel.exists_by([TestModel.data == '1', TestModel.data == '2'])
    assert res is False

    res = await TestModel.exists_by([(TestModel.data == '1') | (TestModel.data == '2')])
    assert res is True

    res = await TestModel.exists_by([(TestModel.data == '1') & (TestModel.id == obj1.id)])
    assert res is True

    res = await TestModel.exists_by([text('data > :data')], params=dict(data='1'))
    assert res is True

    res = await TestModel.exists_by([text('data > :data')], params=dict(data='3'))
    assert res is False

    assert _construct_counter == 3
    assert _build_rela_counter == _construct_counter
    assert _construct_list_counter == 0
    assert _build_rela_list_counter == _construct_list_counter


@pytest.mark.asyncio
async def test_delete(TestModel, build_calls_counter, model_session):
    obj = await TestModel.create(dict(data='123'))

    objects = await TestModel.list()
    assert len(objects) == 1
    assert objects[0] == obj
    obj2 = await TestModel.get(obj.id)
    assert obj == obj2

    assert (await obj.delete()) is True

    objects = await TestModel.list()
    assert len(objects) == 0
    obj2 = await TestModel.get(obj.id)
    assert obj2 is None

    assert (await obj.delete()) is False

    assert _construct_counter == (3 if TestModel.has_cache() else 2)
    assert _build_rela_counter == (2 if TestModel.has_cache() else 1)
    assert _construct_list_counter == 2
    assert _build_rela_list_counter == 1


@pytest.mark.asyncio
async def test_delete_by_id(TestModel, build_calls_counter, model_session):
    obj = await TestModel.create(dict(data='123'))

    objects = await TestModel.list()
    assert len(objects) == 1
    assert objects[0] == obj
    obj2 = await TestModel.get(obj.id)
    assert obj == obj2

    assert (await TestModel.delete_by_id(obj.id)) is True

    objects = await TestModel.list()
    assert len(objects) == 0
    obj2 = await TestModel.get(obj.id)
    assert obj2 is None

    assert (await TestModel.delete_by_id(obj.id)) is False
    assert (await TestModel.delete_by_id(-1)) is False

    assert _construct_counter == (3 if TestModel.has_cache() else 2)
    assert _build_rela_counter == (2 if TestModel.has_cache() else 1)
    assert _construct_list_counter == 2
    assert _build_rela_list_counter == 1


@pytest.mark.asyncio
async def test_delete_by(TestModel, build_calls_counter, model_session):
    obj1 = await TestModel.create(dict(data='1'))
    obj2 = await TestModel.create(dict(data='2'))
    obj3 = await TestModel.create(dict(data='3'))

    res = await TestModel.delete_by([TestModel.data != '1'])
    assert res == 2

    res = await TestModel.delete_by([TestModel.data != '1'])
    assert res == 0

    assert await obj1.load()
    assert not await obj2.load()
    assert not await obj3.load()

    assert _construct_counter == 3
    assert _build_rela_counter == _construct_counter
    assert _construct_list_counter == 0
    assert _build_rela_list_counter == _construct_list_counter


@pytest.mark.asyncio
async def test_save_and_load(TestModel, build_calls_counter, model_session):
    m = TestModel()
    m.id = randint(1, 1000000)
    m.data = 'qwer'
    assert m.create_date is None

    assert not await TestModel.get(m.id)

    assert await m.save()
    assert isinstance(m.create_date, datetime)

    m2 = await TestModel.get(m.id)
    assert m2.id == m.id
    assert m2.data == m.data
    assert isinstance(m2.create_date, datetime)
    assert m2.create_date == m.create_date

    m2.id = 1234
    m2.data = 'qwerty'
    m2.create_date = datetime.now()
    assert await m2.save(updated_fields=['id', 'create_date'])

    m3 = await TestModel.get(1234)
    assert m3.id == 1234
    assert m3.data == 'qwerty'
    assert m3.create_date.replace(microsecond=0) == m2.create_date.replace(microsecond=0)

    assert await m2.load()
    assert m2.id == 1234
    assert m2.data == 'qwerty'

    assert _construct_counter == 2
    assert _build_rela_counter == _construct_counter
    assert _construct_list_counter == 0
    assert _build_rela_list_counter == _construct_list_counter


@pytest.mark.asyncio
async def test_save_all(TestModel, build_calls_counter, model_session):
    obj1 = TestModel(data='1')
    obj2 = TestModel(data='2')
    obj3 = TestModel(data='3')

    assert await TestModel.save_all([obj1, obj2, obj3])

    objects = await TestModel.get_any([obj1.id, obj2.id, obj3.id])
    assert len(objects) == 3
    assert objects[0] == obj1
    assert objects[1] == obj2
    assert objects[2] == obj3
    assert objects[0].id == obj1.id
    assert objects[0].data == obj1.data
    assert objects[1].id == obj2.id
    assert objects[1].data == obj2.data
    assert objects[2].id == obj3.id
    assert objects[2].data == obj3.data

    obj1.data = '4'
    obj2.data = '5'
    obj3.data = '6'

    assert await TestModel.save_all([obj1, obj2, obj3])

    objects = await TestModel.get_any([obj1.id, obj2.id, obj3.id])
    assert len(objects) == 3
    assert objects[0] == obj1
    assert objects[1] == obj2
    assert objects[2] == obj3
    assert objects[0].id == obj1.id
    assert objects[0].data == obj1.data
    assert objects[1].id == obj2.id
    assert objects[1].data == obj2.data
    assert objects[2].id == obj3.id
    assert objects[2].data == obj3.data

    assert _construct_counter == (6 if TestModel.has_cache() else 0)
    assert _build_rela_counter == 0
    assert _construct_list_counter == (2 if TestModel.has_cache() else 0)
    assert _build_rela_list_counter == _construct_list_counter



import pytest

from tests.sql_models_data import USE_SQL

pytestmark = pytest.mark.skipif(not USE_SQL, reason='For SqlAlchemy only')

if USE_SQL:
    from dropland.blocks.sql.base import SqlStorageType
    from dropland.blocks.sql.containers import MultipleSqlStorage, SingleSqlStorage, SqlStorage, SqlBlock
    from dropland.blocks.sql.engine import EngineConfig
    from tests import SQLITE_URI


def test_sync_session_connections(test_sql_storage, sql_sync_engine, sync_sql_session):
    assert sync_sql_session.get_session('dropland-test-sync')
    assert not sync_sql_session.get_session('_')

    for session in sync_sql_session.get_sessions_for_type(SqlStorageType.SQLITE, False):
        assert session.engine
        assert session.engine.db_type == SqlStorageType.SQLITE
        assert session.engine.is_async is False
        assert session.connection
        assert session.in_transaction

    created, session = sync_sql_session.get_or_create_session('dropland-test-sync')
    assert not created
    assert session
    assert session.engine is sql_sync_engine
    assert session.connection
    assert session.in_transaction

    assert not test_sql_storage.get_session('_')
    assert test_sql_storage.get_session('dropland-test-sync') is session


@pytest.mark.asyncio
async def test_async_session_connections(test_sql_storage, sql_async_engine, async_sql_session):
    assert async_sql_session.get_session('dropland-test-async')
    assert not async_sql_session.get_session('_')

    for session in async_sql_session.get_sessions_for_type(SqlStorageType.SQLITE, True):
        assert session.engine
        assert session.engine.db_type == SqlStorageType.SQLITE
        assert session.engine.is_async is True
        assert session.connection
        assert session.in_transaction

    created, session = async_sql_session.get_or_create_session('dropland-test-async')
    assert not created
    assert session
    assert session.engine is sql_async_engine
    assert session.connection
    assert session.in_transaction

    assert not test_sql_storage.get_session('_')
    assert test_sql_storage.get_session('dropland-test-async') is session


def test_sync_container_sessions():
    sqlite_config = EngineConfig(
        url=SQLITE_URI,
        pool_min_size=1, pool_max_size=2,
        pool_expire_seconds=15, pool_timeout_seconds=15,
    )

    cont = SqlStorage()
    engine = cont.create_engine('dropland', sqlite_config, SqlStorageType.SQLITE, use_async=False)
    block = SqlBlock(cont)
    block.sync_startup()
    block.sync_session_begin()

    session = cont.get_session('dropland')
    assert session
    assert session.engine is engine
    assert session.connection
    assert not cont.get_session('_')
    block.sync_session_finish()
    block.sync_shutdown()
    cont.unwire()

    cont = SingleSqlStorage()
    cont.config.from_dict({
        'name': 'dropland',
        'db_type': SqlStorageType.SQLITE,
        'engine_config': sqlite_config,
        'use_async': False,
    })
    engine = cont.create_engine()
    block = SqlBlock(cont)
    block.sync_startup()
    block.sync_session_begin()

    session = cont.get_session()
    assert session
    assert session.engine is engine
    assert session.connection
    block.sync_session_finish()
    block.sync_shutdown()
    cont.unwire()

    cont = MultipleSqlStorage()
    cont.config.from_dict({
        'one': {
            'db_type': SqlStorageType.SQLITE,
            'engine_config': sqlite_config,
            'use_async': False,
        },
        'two': {
            'db_type': SqlStorageType.SQLITE,
            'engine_config': sqlite_config,
            'use_async': False,
        },
    })
    engine1 = cont.create_engine('one')
    engine2 = cont.create_engine('two')
    block = SqlBlock(cont)
    block.sync_startup()
    block.sync_session_begin()

    for name, engine in zip(('one', 'two'), (engine1, engine2)):
        session = cont.get_session(name)
        assert session
        assert session.engine is engine
        assert session.connection

    assert not cont.get_session('_')
    block.sync_session_finish()
    block.sync_shutdown()
    cont.unwire()


@pytest.mark.asyncio
async def test_async_container_sessions():
    sqlite_config = EngineConfig(
        url=SQLITE_URI,
        pool_min_size=1, pool_max_size=2,
        pool_expire_seconds=15, pool_timeout_seconds=15,
    )

    cont = SqlStorage()
    engine = cont.create_engine('dropland', sqlite_config, SqlStorageType.SQLITE, use_async=True)
    block = SqlBlock(cont)
    await block.startup()
    await block.session_begin()

    session = cont.get_session('dropland')
    assert session
    assert session.engine is engine
    assert session.connection
    assert not cont.get_session('_')
    await block.session_finish()
    await block.shutdown()
    cont.unwire()

    cont = SingleSqlStorage()
    cont.config.from_dict({
        'name': 'dropland',
        'db_type': SqlStorageType.SQLITE,
        'engine_config': sqlite_config,
        'use_async': True,
    })
    engine = cont.create_engine()
    block = SqlBlock(cont)
    await block.startup()
    await block.session_begin()

    session = cont.get_session()
    assert session
    assert session.engine is engine
    assert session.connection
    await block.session_finish()
    await block.shutdown()
    cont.unwire()

    cont = MultipleSqlStorage()
    cont.config.from_dict({
        'one': {
            'db_type': SqlStorageType.SQLITE,
            'engine_config': sqlite_config,
            'use_async': True,
        },
        'two': {
            'db_type': SqlStorageType.SQLITE,
            'engine_config': sqlite_config,
            'use_async': True,
        },
    })
    engine1 = cont.create_engine('one')
    engine2 = cont.create_engine('two')
    block = SqlBlock(cont)
    await block.startup()
    await block.session_begin()

    for name, engine in zip(('one', 'two'), (engine1, engine2)):
        session = cont.get_session(name)
        assert session
        assert session.engine is engine
        assert session.connection

    assert not cont.get_session('_')
    await block.session_finish()
    await block.shutdown()
    cont.unwire()



try:
    import starlette

    from .application import FastAPIApplication
    from .model import ApiModelMixin
    from .permissions import HttpPermission, PermDepends, Authenticated, Admin

except ImportError:
    pass



import aio_pika
import pytest

from dropland.blocks.rmq import USE_RMQ

pytestmark = pytest.mark.skipif(not USE_RMQ, reason='For RabbitMQ only')

if USE_RMQ:
    from dropland.blocks.rmq.containers import RmqStorage, SingleRmqStorage, MultipleRmqStorage
    from dropland.blocks.rmq.engine import EngineConfig, RmqStorageEngine
    from tests import RMQ_URI


@pytest.mark.asyncio
async def test_create_engine():
    default_rmq_storage = RmqStorage()
    engine_factory = default_rmq_storage.engine_factory()

    assert not engine_factory.get_engine('')

    config = EngineConfig(url=RMQ_URI)
    assert engine_factory.create_engine('dropland', config)

    engine = engine_factory.get_engine('dropland')
    assert engine
    assert engine.backend is engine_factory
    assert engine.is_async

    assert engine_factory.get_engine_names() == ['dropland']


@pytest.mark.asyncio
async def test_create_connection():
    default_rmq_storage = RmqStorage()
    engine_factory = default_rmq_storage.engine_factory()

    config = EngineConfig(url=RMQ_URI)
    engine = engine_factory.create_engine('dropland', config)

    assert engine
    await engine.async_start()

    async with engine.new_connection() as channel:
        queue = await channel.declare_queue('test', auto_delete=True)
        await channel.default_exchange.publish(
            aio_pika.Message(body=b'12345'),
            routing_key='test',
        )
        async with queue.iterator() as queue_iter:
            async for message in queue_iter:
                async with message.process():
                    assert message.body == b'12345'
                break

    await engine.async_stop()


@pytest.mark.asyncio
async def test_session_connections():
    config = EngineConfig(url=RMQ_URI)

    rmq_storage = SingleRmqStorage()
    rmq_storage.config.from_dict({
        'name': 'dropland1',
        'engine_config': config
    })

    assert rmq_storage.create_engine()
    manager = await rmq_storage.session_context.init()
    session = rmq_storage.get_session()
    assert rmq_storage.manager().get_session('dropland1') == session == manager.get_session('dropland1')


@pytest.mark.asyncio
async def test_storage_container():
    config = EngineConfig(url=RMQ_URI)

    cont = RmqStorage()

    eng = cont.create_engine('dropland', config)
    assert isinstance(eng, RmqStorageEngine)
    assert eng.name == 'dropland'
    cont.unwire()

    cont = SingleRmqStorage()
    cont.config.from_dict({
        'name': '_',
        'engine_config': config
    })
    eng1 = cont.create_engine()
    eng2 = cont.create_engine()
    assert eng1.name == eng2.name == '_'
    assert eng1 is eng2
    cont.unwire()

    cont = MultipleRmqStorage()
    cont.config.from_dict({
        'one': {
            'engine_config': config
        },
        'two': {
            'engine_config': config
        },
    })

    eng1 = cont.create_engine('one')
    eng2 = cont.create_engine('two')
    assert eng1 is not eng2
    assert eng1.is_async is True
    assert eng2.is_async is True
    assert eng1.name == 'one'
    assert eng2.name == 'two'

    cont.unwire()



import asyncio
import contextlib
from dataclasses import dataclass
from datetime import timedelta
from typing import Dict, List, Mapping, Optional

from aioredis import ConnectionsPool, Redis, create_pool

from dropland.log import logger, tr
from ..base import StorageBackend, StorageEngine


@dataclass
class EngineConfig:
    url: str
    max_connections: int = 4
    pool_timeout_seconds: int = 5


class RedisStorageBackend(StorageBackend):
    def __init__(self):
        self._engines: Dict[str, 'RedisStorageEngine'] = dict()

    @property
    def name(self) -> str:
        return 'redis'

    def create_engine(self, name: str, config: EngineConfig,
                      default_ttl: timedelta = timedelta(seconds=60)) -> Optional['RedisStorageEngine']:
        if engine := self._engines.get(name):
            return engine

        engine = RedisStorageEngine(self, name, config, default_ttl)
        self._engines[name] = engine
        logger.info(tr('dropland.blocks.redis.engine.created').format(name=name))
        return engine

    def get_engine(self, name: str) -> Optional['RedisStorageEngine']:
        return self._engines.get(name)

    def get_engine_names(self) -> List[str]:
        return list(self._engines.keys())

    def get_engines(self, names: Optional[List[str]] = None) -> Mapping[str, 'RedisStorageEngine']:
        engines = dict()

        if not names:
            names = self.get_engine_names()

        for name in names:
            if engine := self.get_engine(name):
                engines[name] = engine

        return engines


class RedisStorageEngine(StorageEngine):
    def __init__(self, backend: RedisStorageBackend, name: str, config: EngineConfig, default_ttl: timedelta):
        super().__init__(backend)
        self._name = name
        self._config = config
        self._default_ttl = default_ttl
        self._lock = asyncio.Lock()
        self._pool: Optional[ConnectionsPool] = None
        self._counter = 0

    @property
    def name(self) -> str:
        return self._name

    @property
    def is_async(self):
        return True

    @property
    def default_ttl(self) -> timedelta:
        return self._default_ttl

    def new_connection(self):
        @contextlib.asynccontextmanager
        async def inner(self):
            yield Redis(self._pool)
        return inner(self)

    def start(self):
        raise RuntimeError('Use async_start method')

    def stop(self):
        raise RuntimeError('Use async_stop method')

    async def async_start(self):
        async with self._lock:
            if 0 == self._counter:
                assert not self._pool
                self._pool = await create_pool(
                    self._config.url, create_connection_timeout=self._config.pool_timeout_seconds,
                    minsize=1, maxsize=self._config.max_connections
                )
                logger.info(tr('dropland.blocks.redis.engine.started'))
            self._counter += 1

    async def async_stop(self):
        async with self._lock:
            if 1 == self._counter:
                assert self._pool
                self._pool.close()
                await self._pool.wait_closed()
                self._pool = None
                logger.info(tr('dropland.blocks.redis.engine.stopped'))

            self._counter = max(self._counter - 1, 0)



try:
    import aioredis

    from .engine import EngineConfig, RedisStorageBackend, RedisStorageEngine
    from .model import RedisModel, RedisMethodCache
    from .session import SessionManager
    from .settings import RedisSettings

    USE_REDIS = True

except ImportError:
    USE_REDIS = False



import asyncio
import contextlib
import sys
import threading
from collections import defaultdict
from dataclasses import dataclass, replace
from datetime import timedelta
from typing import Dict, List, Mapping, Optional, Set

import ujson
from sqlalchemy import create_engine as create_sync_engine
from sqlalchemy.ext.asyncio import AsyncSession, create_async_engine
from sqlalchemy.orm import Session, sessionmaker
from sqlalchemy.schema import MetaData

from dropland.blocks.sql.base import SqlStorageType
from dropland.log import logger, tr
from .base import NAMING_CONVENTION
from ..base import StorageBackend, StorageEngine


@dataclass
class EngineConfig:
    url: str
    echo: bool = False
    pool_min_size: int = 1
    pool_max_size: int = 8
    pool_expire_seconds: int = 60
    pool_timeout_seconds: int = 15


@dataclass
class EngineKey:
    db_type: SqlStorageType
    is_async: bool

    def __eq__(self, other: 'EngineKey'):
        return (self.db_type, self.is_async) == (other.db_type, other.is_async)

    def __hash__(self):
        return hash((self.db_type, self.is_async))


@dataclass
class EngineWithDbType:
    db_type: SqlStorageType
    engine: 'SqlStorageEngine'


class SqlStorageBackend(StorageBackend):
    def __init__(self):
        self._engines: Dict[str, 'SqlStorageEngine'] = dict()
        self._engines_by_key: Dict[EngineKey, Set[str]] = defaultdict(set)
        self._engines_by_type: Dict[SqlStorageType, Set[str]] = defaultdict(set)

    @property
    def name(self) -> str:
        return 'sqla'

    def create_engine(self, name: str, config: EngineConfig, db_type: SqlStorageType, use_async: bool) \
            -> Optional['SqlStorageEngine']:
        if engine := self._engines.get(name):
            return engine

        uri = config.url

        if use_async:
            if db_type == SqlStorageType.SQLITE:
                uri = uri.replace('sqlite', 'sqlite+aiosqlite', 1)
            elif db_type == SqlStorageType.POSTGRES:
                uri = uri.replace('postgresql', 'postgresql+asyncpg', 1)
            elif db_type == SqlStorageType.MYSQL:
                uri = uri.replace('mysql', 'mysql+aiomysql', 1)
        else:
            if db_type == SqlStorageType.MYSQL:
                uri = uri.replace('mysql', 'mysql+pymysql', 1)

        config = replace(config, url=uri)

        params = dict(
            echo=config.echo,
            echo_pool=config.echo,
            pool_pre_ping=True,
            pool_recycle=config.pool_expire_seconds,
            json_serializer=ujson.dumps,
            json_deserializer=ujson.loads,
        )

        if db_type in (SqlStorageType.POSTGRES, SqlStorageType.MYSQL):
            params.update(dict(
                pool_size=config.pool_min_size,
                max_overflow=config.pool_max_size - config.pool_min_size,
                pool_timeout=config.pool_timeout_seconds
            ))

        if use_async:
            engine = create_async_engine(config.url, **params)
            session_class = sessionmaker(engine, autoflush=False, expire_on_commit=False, class_=AsyncSession)
        else:
            engine = create_sync_engine(config.url, **params)
            session_class = sessionmaker(engine, autoflush=False)

        logger.info(tr('dropland.blocks.sql.engine.created').format(db_type=db_type, async_=f'async={use_async}'))
        metadata = MetaData(bind=engine, naming_convention=NAMING_CONVENTION)
        engine_class = SqlAlchemyAsyncEngine if use_async else SqlAlchemySyncEngine
        engine = engine_class(self, name, engine, metadata, db_type, session_class,
                              timedelta(seconds=config.pool_timeout_seconds))

        self._engines[name] = engine
        self._engines_by_key[EngineKey(db_type=db_type, is_async=use_async)].add(name)
        self._engines_by_type[db_type].add(name)
        return engine

    def get_engine(self, name: str) -> Optional['SqlStorageEngine']:
        return self._engines.get(name)

    def get_engine_names(self) -> List[str]:
        return list(self._engines.keys())

    def get_engines_for_type(self, db_type: SqlStorageType, is_async: bool) -> List['SqlStorageEngine']:
        engine_names = self._engines_by_key[EngineKey(db_type=db_type, is_async=is_async)]
        return [self.get_engine(name) for name in engine_names]

    def get_engines(self, names: Optional[List[str]] = None) -> Mapping[str, 'SqlStorageEngine']:
        engines = dict()

        if not names:
            names = self.get_engine_names()

        for name in names:
            if engine := self.get_engine(name):
                engines[name] = engine

        return engines


class SqlStorageEngine(StorageEngine):
    def __init__(self, backend: SqlStorageBackend, name: str, raw_engine, metadata,
                 db_type: SqlStorageType, timeout: timedelta):
        super().__init__(backend)
        self._name = name
        self._engine = raw_engine
        self._db_type = db_type
        self._timeout = timeout
        self._metadata = metadata

    @property
    def name(self) -> str:
        return self._name

    @property
    def raw_engine(self):
        return self._engine

    @property
    def db_type(self):
        return self._db_type

    @property
    def timeout(self) -> timedelta:
        return self._timeout

    @property
    def metadata(self):
        return self._metadata

    @property
    def is_async(self):
        raise NotImplementedError

    @property
    def connection_class(self):
        raise NotImplementedError

    def new_connection(self):
        raise NotImplementedError

    def sync_start(self):
        raise NotImplementedError

    def sync_stop(self):
        raise NotImplementedError

    async def start(self):
        raise NotImplementedError

    async def stop(self):
        raise NotImplementedError

    @contextlib.contextmanager
    def sync_transaction_context(self, connection, autocommit: bool = True):
        raise NotImplementedError

    @contextlib.asynccontextmanager
    async def transaction_context(self, connection, autocommit: bool = True):
        raise NotImplementedError


# noinspection PyPep8Naming
class SqlAlchemySyncEngine(SqlStorageEngine):
    def __init__(self, backend: SqlStorageBackend, name: str, engine, metadata,
                 db_type: SqlStorageType, session_class, timeout: timedelta):
        super().__init__(backend, name, engine, metadata, db_type, timeout)
        self._session_class = session_class
        self._lock = threading.Lock()
        self._counter = 0

    @property
    def is_async(self):
        return False

    @property
    def connection_class(self):
        return Session

    def new_connection(self):
        return self._session_class()

    def sync_start(self):
        with self._lock:
            if 0 == self._counter:
                logger.info(tr('dropland.blocks.sql.engine.started')
                            .format(db_type=self.db_type, async_=f'async=False'))
            self._counter += 1

    def sync_stop(self):
        with self._lock:
            if 1 == self._counter:
                self.raw_engine.dispose()
                logger.info(tr('dropland.blocks.sql.engine.stopped')
                            .format(db_type=self.db_type, async_=f'async=False'))
            self._counter = max(self._counter - 1, 0)

    async def start(self):
        raise RuntimeError('Use start method')

    async def stop(self):
        raise RuntimeError('Use stop method')

    @contextlib.contextmanager
    def sync_transaction_context(self, connection: Session, autocommit: bool = True):
        with connection.begin() as tx:
            tx = connection

            yield tx

            if sys.exc_info()[0]:
                tx.rollback()
            else:
                if autocommit:
                    tx.commit()
                else:
                    tx.rollback()

    @contextlib.asynccontextmanager
    async def transaction_context(self, connection, autocommit: bool = True):
        raise RuntimeError('Use sync_transaction_context method')


# noinspection PyPep8Naming
class SqlAlchemyAsyncEngine(SqlStorageEngine):
    def __init__(self, backend: SqlStorageBackend, name: str, engine, metadata,
                 db_type: SqlStorageType, session_class, timeout: timedelta):
        super().__init__(backend, name, engine, metadata, db_type, timeout)
        self._session_class = session_class
        self._lock = asyncio.Lock()
        self._counter = 0

    @property
    def is_async(self):
        return True

    @property
    def connection_class(self):
        return AsyncSession

    def new_connection(self):
        return self._session_class()

    def sync_start(self):
        raise RuntimeError('Use start method')

    def sync_stop(self):
        raise RuntimeError('Use stop method')

    async def start(self):
        async with self._lock:
            if 0 == self._counter:
                logger.info(tr('dropland.blocks.sql.engine.started')
                            .format(db_type=self.db_type, async_=f'async=True'))
            self._counter += 1

    async def stop(self):
        async with self._lock:
            if 1 == self._counter:
                await self.raw_engine.dispose()
                logger.info(tr('dropland.blocks.sql.engine.stopped')
                            .format(db_type=self.db_type, async_=f'async=True'))
            self._counter = max(self._counter - 1, 0)

    @contextlib.contextmanager
    def sync_transaction_context(self, connection: Session, autocommit: bool = True):
        raise RuntimeError('Use transaction_context method')

    @contextlib.asynccontextmanager
    async def transaction_context(self, connection: AsyncSession, autocommit: bool = True):
        async with connection.begin() as tx:
            yield tx

            if sys.exc_info()[0]:
                await tx.rollback()
            else:
                if autocommit:
                    await tx.commit()
                else:
                    await tx.rollback()



import contextlib

from dependency_injector import containers, providers

from dropland.app.base import ContainerBlock, SessionResourceBlock
from dropland.data.context import get_context
from dropland.util import default_value
from .engine import EngineConfig, SqlStorageBackend
from .session import SessionManager


class SqlBlock(ContainerBlock, SessionResourceBlock):
    def sync_startup(self, *args, **kwargs):
        if self.initialized:
            return
        self.container.init_resources()
        self._initialized = True

    def sync_shutdown(self, *args, **kwargs):
        if not self.initialized:
            return
        self.container.shutdown_resources()
        self._initialized = False

    async def startup(self, *args, **kwargs):
        if self.initialized:
            return
        await self.container.init_resources()
        self._initialized = True

    async def shutdown(self, *args, **kwargs):
        if not self.initialized:
            return
        await self.container.shutdown_resources()
        self._initialized = False

    def sync_session_begin(self, begin_tx: bool = True, autocommit: bool = True, *args, **kwargs):
        ctx = get_context()
        if not self.initialized or '_sql_sync_block' in ctx:
            return

        ctx._sql_sync_block = contextlib.ExitStack().__enter__()
        for name, engine in self.container.engine_factory().get_engines().items():
            if not engine.is_async:
                ctx._sql_sync_block.enter_context(
                    self.container.manager().sync_session_context(name, begin_tx=begin_tx, autocommit=autocommit)
                )

    def sync_session_finish(self, *args, **kwargs):
        ctx = get_context()
        if not self.initialized or '_sql_sync_block' not in ctx:
            return

        ctx._sql_sync_block.__exit__(None, None, None)
        del ctx._sql_sync_block

    async def session_begin(self, begin_tx: bool = True, autocommit: bool = True, *args, **kwargs):
        ctx = get_context()
        if not self.initialized or '_sql_block' in ctx:
            return

        ctx._sql_block = await contextlib.AsyncExitStack().__aenter__()
        for name, engine in self.container.engine_factory().get_engines().items():
            if engine.is_async:
                await ctx._sql_block.enter_async_context(
                    self.container.manager().session_context(name, begin_tx=begin_tx, autocommit=autocommit)
                )

    async def session_finish(self, *args, **kwargs):
        ctx = get_context()
        if not self.initialized or '_sql_block' not in ctx:
            return

        await ctx._sql_block.__aexit__(None, None, None)
        del ctx._sql_block

    def get_engine(self, *args, **kwargs):
        if isinstance(self.container, SingleSqlStorage) or 'SingleSqlStorage' == self.container.parent_name:
            return self.container.create_engine()
        return self.container.create_engine(*args, **kwargs)

    def get_session(self, *args, **kwargs):
        if isinstance(self.container, SingleSqlStorage) or 'SingleSqlStorage' == self.container.parent_name:
            return self.container.get_session()
        return self.container.get_session(*args, **kwargs)


class SqlStorage(containers.DeclarativeContainer):
    __self__ = providers.Self()
    engine_factory = providers.Singleton(SqlStorageBackend)
    manager = providers.Singleton(SessionManager, engine_factory)

    def _create_engine(self, *args, **kwargs):
        return self.engine_factory().create_engine(*args, **kwargs)

    def _init_sync_session(self):
        with self.manager().init_sync_engines():
            yield self.manager()

    async def _init_session(self):
        async with self.manager().init_engines():
            yield self.manager()

    create_engine = providers.Factory(_create_engine, __self__)
    sync_session_context = providers.Resource(_init_sync_session, __self__)
    session_context = providers.Resource(_init_session, __self__)
    get_session = providers.Factory(manager.provided.get_session.call())
    block = providers.Singleton(SqlBlock, __self__)

    wiring_config = containers.WiringConfiguration(
        modules=['.model']
    )


class SingleSqlStorage(SqlStorage):
    __self__ = providers.Self()
    config = providers.Configuration()

    def _create_engine(self):
        if isinstance(self.config.engine_config(), EngineConfig):
            engine_config = self.config.engine_config()
        else:
            engine_config = EngineConfig(
                url=self.config.engine_config.url(),
                echo=self.config.engine_config.echo.as_(bool)(),
                pool_min_size=self.config.engine_config.
                    pool_min_size.as_(default_value(int))(default=1),
                pool_max_size=self.config.engine_config.
                    pool_max_size.as_(default_value(int))(default=8),
                pool_expire_seconds=self.config.engine_config.
                    pool_expire_seconds.as_(default_value(int))(default=60),
                pool_timeout_seconds=self.config.engine_config.
                    pool_timeout_seconds.as_(default_value(int))(default=15)
            )
        return SqlStorage._create_engine(
            self, self.config.name(), engine_config, self.config.db_type(),
            use_async=self.config.use_async.as_(bool)())

    create_engine = providers.Factory(_create_engine, __self__)
    sync_session_context = providers.Resource(SqlStorage._init_sync_session, __self__)
    session_context = providers.Resource(SqlStorage._init_session, __self__)
    get_session = providers.Factory(SqlStorage.get_session, config.name)
    block = providers.Singleton(SqlBlock, __self__)

    wiring_config = containers.WiringConfiguration(
        modules=['.model']
    )


class MultipleSqlStorage(SqlStorage):
    __self__ = providers.Self()
    config = providers.Configuration()

    def _create_engine(self, name: str):
        if conf := self.config.get(name):
            if isinstance(conf['engine_config'], EngineConfig):
                engine_config = conf['engine_config']
            else:
                engine_config = EngineConfig(
                    url=conf['engine_config']['url'],
                    echo=bool(conf['engine_config'].get('echo')),
                    pool_min_size=int(conf['engine_config'].get('pool_min_size', 1)),
                    pool_max_size=int(conf['engine_config'].get('pool_max_size', 8)),
                    pool_expire_seconds=int(conf['engine_config'].get('pool_expire_seconds', 60)),
                    pool_timeout_seconds=int(conf['engine_config'].get('pool_timeout_seconds', 15))
                )
            return SqlStorage._create_engine(
                self, name, engine_config, conf['db_type'], use_async=conf.get('use_async', False))
        return None

    def _get_session(self, name: str):
        return self.manager().get_session(name)

    create_engine = providers.Factory(_create_engine, __self__)
    sync_session_context = providers.Resource(SqlStorage._init_sync_session, __self__)
    session_context = providers.Resource(SqlStorage._init_session, __self__)
    get_session = providers.Factory(_get_session, __self__)
    block = providers.Singleton(SqlBlock, __self__)

    wiring_config = containers.WiringConfiguration(
        modules=['.model']
    )



try:
    import aioredis

    from .engine import EngineConfig, SchedulerBackend
    from .local import Scheduler
    from .application import SchedulerApplication, SchedulerBlock, SchedulerRPC
    from .settings import SchedulerSettings

    USE_SCHEDULER = True

except ImportError:
    USE_SCHEDULER = False



class StorageBackend:
    @property
    def name(self) -> str:
        raise NotImplementedError


class StorageEngine:
    def __init__(self, backend: StorageBackend):
        self._backend = backend

    @property
    def backend(self):
        return self._backend

    @property
    def is_async(self):
        raise NotImplementedError

    def new_connection(self):
        raise NotImplementedError

    def start(self):
        pass

    def stop(self):
        pass

    async def async_start(self):
        pass

    async def async_stop(self):
        pass



import contextlib
from datetime import timedelta

from dependency_injector import containers, providers
from pytimeparse.timeparse import timeparse

from dropland.app.base import ContainerBlock, SessionResourceBlock
from dropland.data.context import get_context
from dropland.util import default_value
from .engine import EngineConfig, RedisStorageBackend
from .session import SessionManager


class RedisBlock(ContainerBlock, SessionResourceBlock):
    async def startup(self, *args, **kwargs):
        if self.initialized:
            return
        await self.container.init_resources()
        self._initialized = True

    async def shutdown(self, *args, **kwargs):
        if not self.initialized:
            return
        await self.container.shutdown_resources()
        self._initialized = False

    async def session_begin(self, *args, **kwargs):
        ctx = get_context()
        if not self.initialized or '_redis_block' in ctx:
            return

        ctx._redis_block = await contextlib.AsyncExitStack().__aenter__()
        for engine_name in self.container.engine_factory().get_engine_names():
            await ctx._redis_block.enter_async_context(
                self.container.manager().session_context(engine_name)
            )

    async def session_finish(self, *args, **kwargs):
        ctx = get_context()
        if not self.initialized or '_redis_block' not in ctx:
            return

        await ctx._redis_block.__aexit__(None, None, None)
        del ctx._redis_block

    def get_engine(self, *args, **kwargs):
        if isinstance(self.container, SingleRedisStorage) or 'SingleRedisStorage' == self.container.parent_name:
            return self.container.create_engine()
        return self.container.create_engine(*args, **kwargs)

    def get_session(self, *args, **kwargs):
        if isinstance(self.container, SingleRedisStorage) or 'SingleRedisStorage' == self.container.parent_name:
            return self.container.get_session()
        return self.container.get_session(*args, **kwargs)


class RedisStorage(containers.DeclarativeContainer):
    __self__ = providers.Self()
    engine_factory = providers.Singleton(RedisStorageBackend)
    manager = providers.Singleton(SessionManager, engine_factory)
    default_ttl = providers.Object(timedelta(seconds=timeparse('1 min')))

    def _create_engine(self, *args, **kwargs):
        return self.engine_factory().create_engine(*args, **kwargs)

    async def _init_session(self):
        async with self.manager().init_engines():
            yield self.manager()

    create_engine = providers.Factory(_create_engine, __self__)
    session_context = providers.Resource(_init_session, __self__)
    get_session = providers.Factory(manager.provided.get_session.call())
    block = providers.Singleton(RedisBlock, __self__)

    wiring_config = containers.WiringConfiguration(
        modules=['.model']
    )


class SingleRedisStorage(RedisStorage):
    __self__ = providers.Self()
    config = providers.Configuration()
    # noinspection PyArgumentList
    default_ttl = providers.Object(timedelta(
        seconds=timeparse(config.get('default_cache_ttl', required=False) or '1 min')))

    def _create_engine(self):
        if isinstance(self.config.engine_config(), EngineConfig):
            engine_config = self.config.engine_config()
        else:
            engine_config = EngineConfig(
                url=self.config.engine_config.url(),
                max_connections=self.config.engine_config.
                    max_connections.as_(default_value(int))(default=4),
                pool_timeout_seconds=self.config.engine_config.
                    pool_timeout_seconds.as_(default_value(int))(default=5)
            )
        return RedisStorage._create_engine(self, self.config.name(), engine_config, self.default_ttl())

    create_engine = providers.Factory(_create_engine, __self__)
    session_context = providers.Resource(RedisStorage._init_session, __self__)
    get_session = providers.Factory(RedisStorage.get_session, config.name)
    block = providers.Singleton(RedisBlock, __self__)

    wiring_config = containers.WiringConfiguration(
        modules=['.model']
    )


class MultipleRedisStorage(RedisStorage):
    __self__ = providers.Self()
    config = providers.Configuration()

    def _create_engine(self, name: str):
        if conf := self.config.get(name):
            if isinstance(conf['engine_config'], EngineConfig):
                engine_config = conf['engine_config']
            else:
                engine_config = EngineConfig(
                    url=conf['engine_config']['url'],
                    max_connections=int(conf['engine_config'].get('max_connections', 4)),
                    pool_timeout_seconds=int(conf['engine_config'].get('pool_timeout_seconds', 5))
                )
            return RedisStorage._create_engine(
                self, name, engine_config,
                timedelta(seconds=timeparse(conf.get('default_cache_ttl', '1 min'))))
        return None

    def _get_session(self, name: str):
        return self.manager().get_session(name)

    create_engine = providers.Factory(_create_engine, __self__)
    session_context = providers.Resource(RedisStorage._init_session, __self__)
    get_session = providers.Factory(_get_session, __self__)
    block = providers.Singleton(RedisBlock, __self__)

    wiring_config = containers.WiringConfiguration(
        modules=['.model']
    )



from typing import Awaitable, Callable, List, Optional

from dependency_injector.containers import Container
from fastapi import FastAPI, Request, Response
from starlette.middleware import Middleware
from starlette.middleware.base import BaseHTTPMiddleware

from dropland.app.application import Application


class FastAPIApplication(Application):
    def __init__(self, container: Container, name: str, debug: bool, version: str,
                 title: str, api_root: str, cors: Optional[List[str]] = None):
        super().__init__(container, name, debug, version)

        self._server = self._create_server(title, debug, self.version, api_root, cors)
        self._server.state.application = self

    @property
    def server(self):
        return self._server

    def _create_server(self, title: str, debug: bool, version: str,
                       api_root: str, cors: Optional[List[str]] = None):
        from fastapi.responses import UJSONResponse
        from starlette.authentication import AuthenticationBackend, AuthCredentials, UnauthenticatedUser
        from starlette.middleware.authentication import AuthenticationMiddleware
        from fastapi.middleware.cors import CORSMiddleware

        class AuthBackend(AuthenticationBackend):
            async def authenticate(self, request):
                return AuthCredentials(), UnauthenticatedUser()

        middlewares = [
            Middleware(
                CORSMiddleware,
                allow_origins=[str(origin) for origin in cors] if cors else ['*'],
                allow_credentials=True, allow_methods=['*'], allow_headers=['*'],
            ),
            Middleware(AuthenticationMiddleware, backend=AuthBackend()),
        ]

        server = FastAPI(
            title=title, version=version, debug=debug,
            default_response_class=UJSONResponse, root_path=api_root,
            on_startup=[self.startup, self.sync_startup],
            on_shutdown=[self.shutdown, self.sync_shutdown],
            middleware=middlewares
        )

        server.add_middleware(BaseHTTPMiddleware, dispatch=self._server_entrypoint)

        return server

    async def _server_entrypoint(self, request: Request, call_next: Callable[[Request], Awaitable[Response]]):
        return await call_next(request)



from pathlib import Path
from typing import Any, Dict, Optional

from pydantic.class_validators import validator
from pydantic.env_settings import BaseSettings
from pydantic.networks import RedisDsn


class RedisSettings(BaseSettings):
    REDIS_TYPE: str = 'redis'
    REDIS_HOST: str = 'localhost'
    REDIS_PORT: int = 6379
    REDIS_PASSWORD: str = ''
    REDIS_PASSWORD_FILE: Optional[Path] = None
    REDIS_DB: int = 0
    REDIS_MAX_CONNECTIONS: int = 4
    REDIS_POOL_TIMEOUT_SECONDS: int = 5

    REDIS_URL: Optional[RedisDsn] = None

    @validator('REDIS_URL', pre=True)
    def assemble_redis_connection(cls, v: Optional[str], values: Dict[str, Any]) -> str:
        if isinstance(v, str):
            return v

        if 'REDIS_PASSWORD_FILE' in values:
            password_file = values.get('REDIS_PASSWORD_FILE')
            if isinstance(password_file, Path) and password_file.exists():
                with password_file.open() as f:
                    values['REDIS_PASSWORD'] = f.read().strip()

        return RedisDsn.build(
            scheme=values.get('REDIS_TYPE'),
            password=values.get('REDIS_PASSWORD'),
            host=values.get('REDIS_HOST'),
            port=str(values.get('REDIS_PORT')),
            path=f'/{values.get("REDIS_DB") or ""}',
        )

    class Config:
        case_sensitive = False
        env_file = '.env'



import contextlib
from typing import Dict, Any

from contextvars import ContextVar


class ContextData(object):
    def __init__(self, props: Dict[str, Any] = None, **kwargs):
        if props:
            for name, value in props.items():
                super().__setattr__(name, self._wrap(value))
        super().__setattr__('data', kwargs)

    def _wrap(self, value):
        if isinstance(value, (tuple, list, set, frozenset)):
            return type(value)([self._wrap(v) for v in value])
        else:
            return value

    def __setattr__(self, key, value):
        super().__setattr__(key, self._wrap(value))

    def __contains__(self, item):
        return item in self.__dict__


_ctx: ContextVar[ContextData] = ContextVar('_ctx')


def get_context():
    if ctx := _ctx.get(None):
        return ctx
    _ctx.set(ContextData())
    return _ctx.get()


@contextlib.contextmanager
def with_context(force_new: bool = False):
    if not force_new:
        if ctx := _ctx.get(None):
            yield ctx
            return

    token = _ctx.set(ContextData())
    yield _ctx.get(None)
    _ctx.reset(token)



import contextlib

from dependency_injector import containers, providers

from dropland.app.base import ContainerBlock, SessionResourceBlock
from dropland.data.context import get_context
from dropland.util import default_value
from .engine import EngineConfig, RmqStorageBackend
from .session import SessionManager


class RmqBlock(ContainerBlock, SessionResourceBlock):
    async def startup(self, *args, **kwargs):
        if self.initialized:
            return
        await self.container.init_resources()
        self._initialized = True

    async def shutdown(self, *args, **kwargs):
        if not self.initialized:
            return
        await self.container.shutdown_resources()
        self._initialized = False

    async def session_begin(self, *args, **kwargs):
        ctx = get_context()
        if not self.initialized or '_rmq_block' in ctx:
            return

        ctx._rmq_block = await contextlib.AsyncExitStack().__aenter__()
        for engine_name in self.container.engine_factory().get_engine_names():
            await ctx._rmq_block.enter_async_context(
                self.container.manager().session_context(engine_name)
            )

    async def session_finish(self, *args, **kwargs):
        ctx = get_context()
        if not self.initialized or '_rmq_block' not in ctx:
            return

        await ctx._rmq_block.__aexit__(None, None, None)
        del ctx._rmq_block

    def get_engine(self, *args, **kwargs):
        if isinstance(self.container, SingleRmqStorage) or 'SingleRmqStorage' == self.container.parent_name:
            return self.container.create_engine()
        return self.container.create_engine(*args, **kwargs)

    def get_session(self, *args, **kwargs):
        if isinstance(self.container, SingleRmqStorage) or 'SingleRmqStorage' == self.container.parent_name:
            return self.container.get_session()
        return self.container.get_session(*args, **kwargs)


class RmqStorage(containers.DeclarativeContainer):
    __self__ = providers.Self()
    engine_factory = providers.Singleton(RmqStorageBackend)
    manager = providers.Singleton(SessionManager, engine_factory)

    def _create_engine(self, *args, **kwargs):
        return self.engine_factory().create_engine(*args, **kwargs)

    async def _init_session(self):
        async with self.manager().init_engines():
            yield self.manager()

    create_engine = providers.Factory(_create_engine, __self__)
    session_context = providers.Resource(_init_session, __self__)
    get_session = providers.Factory(manager.provided.get_session.call())
    block = providers.Singleton(RmqBlock, __self__)

    wiring_config = containers.WiringConfiguration(
        modules=[__name__]
    )


class SingleRmqStorage(RmqStorage):
    __self__ = providers.Self()
    config = providers.Configuration()

    def _create_engine(self):
        if isinstance(self.config.engine_config(), EngineConfig):
            engine_config = self.config.engine_config()
        else:
            engine_config = EngineConfig(
                url=self.config.engine_config.url(),
                virtualhost=self.config.engine_config.virtualhost(),
                timeout_seconds=self.config.engine_config.
                    timeout_seconds.as_(default_value(int))(default=5),
                pool_max_connections=self.config.engine_config.
                    pool_max_connections.as_(default_value(int))(default=4),
                pool_max_channels_per_connection=self.config.engine_config.
                    pool_max_channels_per_connection.as_(default_value(int))(default=100)
            )
        return RmqStorage._create_engine(self, self.config.name(), engine_config)

    create_engine = providers.Factory(_create_engine, __self__)
    session_context = providers.Resource(RmqStorage._init_session, __self__)
    get_session = providers.Factory(RmqStorage.get_session, config.name)
    block = providers.Singleton(RmqBlock, __self__)

    wiring_config = containers.WiringConfiguration(
        modules=[__name__]
    )


class MultipleRmqStorage(RmqStorage):
    __self__ = providers.Self()
    config = providers.Configuration()

    def _create_engine(self, name: str):
        if conf := self.config.get(name):
            if isinstance(conf['engine_config'], EngineConfig):
                engine_config = conf['engine_config']
            else:
                engine_config = EngineConfig(
                    url=conf['engine_config']['url'],
                    virtualhost=conf['engine_config']['virtualhost'],
                    timeout_seconds=int(conf['engine_config'].get('timeout_seconds', 5)),
                    pool_max_connections=int(conf['engine_config'].get('pool_max_connections', 4)),
                    pool_max_channels_per_connection=int(conf['engine_config'].get(
                        'pool_max_channels_per_connection', 100))
                )
            return RmqStorage._create_engine(self, name, engine_config)
        return None

    def _get_session(self, name: str):
        return self.manager().get_session(name)

    create_engine = providers.Factory(_create_engine, __self__)
    session_context = providers.Resource(RmqStorage._init_session, __self__)
    get_session = providers.Factory(_get_session, __self__)
    block = providers.Singleton(RmqBlock, __self__)

    wiring_config = containers.WiringConfiguration(
        modules=[__name__]
    )



from dataclasses import dataclass
from datetime import timedelta
from typing import Protocol, Any, List, Optional, Dict, Tuple, AsyncIterable

from .context import ContextData


@dataclass
class ModelCacheData:
    cache_id: str
    data: Optional[Any] = None


class ModelCacheProtocol(Protocol):
    def get_model_cache_key(self) -> str:
        ...

    def get_cache_id(self, id_value: Any) -> str:
        return str(id_value)

    def get_cache_key(self, id_value: Any) -> str:
        return f'{self.get_model_cache_key()}:{self.get_cache_id(id_value)}'

    async def cache_one(
            self, ctx: ContextData, data: ModelCacheData, ttl: Optional[timedelta] = None, **kwargs) -> bool:
        ...

    async def cache_many(
            self, ctx: ContextData, objects: List[ModelCacheData], ttl: Optional[timedelta] = None, **kwargs) -> bool:
        ...

    async def load_one(
            self, ctx: ContextData, cache_key: str, **kwargs) -> Tuple[bool, Optional[Any]]:
        ...

    async def load_many(
            self, ctx: ContextData, indices: List[Any], **kwargs) -> List[Optional[Dict[str, Any]]]:
        ...

    async def drop_one(self, ctx: ContextData, cache_key: str) -> bool:
        ...

    async def drop_many(self, ctx: ContextData, indices: List[Any] = None) -> int:
        ...

    async def drop_all(self, ctx: ContextData, prefix: Optional[str] = None) -> int:
        ...

    async def exists(self, ctx: ContextData, cache_key: str) -> bool:
        ...

    async def scan(self, ctx: ContextData, cache_key: str = None,
                   match: str = None, count: int = None) -> AsyncIterable[Tuple[str, Optional[Dict[str, Any]]]]:
        ...



import inspect
import sys
from importlib.abc import MetaPathFinder, Loader
from importlib.machinery import ModuleSpec
from importlib.metadata import entry_points
from importlib.util import find_spec


class ProxyLoader(Loader):
    def __init__(self, fullname, loader):
        self._fullname = fullname
        self._loader = loader

    def create_module(self, spec):
        return self._loader.create_module(spec)

    def exec_module(self, mod):
        sys.modules[self._fullname] = mod
        return self._loader.exec_module(mod)


class NoopLoader(Loader):
    def __init__(self, mod):
        self._mod = mod

    def create_module(self, spec):
        return self._mod

    def exec_module(self, mod):
        pass


class ModuleLoader(MetaPathFinder):
    def __init__(self, ep_name: str):
        caller = inspect.currentframe().f_back
        self._module_name = caller.f_globals['__name__']

        self._redirects = {
            f'{self._module_name}.{ep.name}': ep.value
            for ep in entry_points().get(ep_name, [])
        }

    # noinspection PyUnusedLocal
    def find_spec(self, fullname, path, target=None):
        target = self._redirects.get(fullname)
        if target:
            mod = sys.modules.get(target)
            if mod is None:
                spec = find_spec(target)
                spec.loader = ProxyLoader(fullname, spec.loader)
                return spec
            else:
                return ModuleSpec(fullname, NoopLoader(mod))
        elif fullname.startswith(self._module_name):
            raise ImportError(
                f'Cannot import {fullname} - is {fullname[len(__name__) + 1:]} a valid and installed?'
            )

    @classmethod
    def uninstall(cls):
        if sys.meta_path:
            for i in range(len(sys.meta_path) - 1, -1, -1):
                if type(sys.meta_path[i]).__name__ == cls.__name__:
                    del sys.meta_path[i]

    def install(self):
        self.uninstall()
        sys.meta_path.append(self)



try:
    from .engine import EngineConfig, SqlStorageBackend, SqlStorageEngine
    from .model import SqlaModelBase as SqlaModel
    from .session import SessionManager
    from .settings import SqliteSettings, PgSettings, MySqlSettings

    USE_SQL = True

except ImportError:
    USE_SQL = False



# -*- coding: utf-8 -*-
import os

from setuptools import setup

setup_kwargs = {
    'dependency_links': [
        os.path.join(os.getcwd(), 'backends', 'sqla'),
        os.path.join(os.getcwd(), 'backends', 'gino'),
    ]
}

setup(**setup_kwargs)



from typing import Protocol, Any, Optional, List, Tuple

from dropland.blocks.base import StorageEngine


class ModelProtocol(Protocol):
    def get_id_value(self) -> Any:
        ...

    @classmethod
    def get_engine(cls) -> StorageEngine:
        ...

    #
    # Retrieve operations
    #

    @classmethod
    async def get(cls, id_value: Any, **kwargs) -> Optional['ModelProtocol']:
        ...

    @classmethod
    async def get_any(cls, indices: List[Any], **kwargs) -> List[Optional['ModelProtocol']]:
        ...

    @classmethod
    async def exists(cls, id_value: Any, **kwargs) -> bool:
        ...

    #
    # Modification operations
    #

    @classmethod
    async def create(cls, data: Any, **kwargs) -> Optional['ModelProtocol']:
        ...

    @classmethod
    async def get_or_create(cls, id_value: Any, data:Any, **kwargs) -> Tuple[Optional['ModelProtocol'], bool]:
        ...

    async def update(self, data: Any) -> bool:
        ...

    @classmethod
    async def update_by_id(cls, id_value: Any, data: Any) -> bool:
        ...

    async def save(self, *args, **kwargs) -> bool:
        ...

    async def load(self, field_names: List[str] = None) -> bool:
        ...

    @classmethod
    async def save_all(cls, objects: List['ModelProtocol'], *args, **kwargs) -> bool:
        ...

    async def delete(self) -> bool:
        ...

    @classmethod
    async def delete_all(cls, indices: List[Any] = None) -> bool:
        ...

    @classmethod
    async def delete_by_id(cls, id_value: Any) -> bool:
        ...



from dropland.blocks.sql import USE_SQL

if USE_SQL:
    from tests.conftest import sql_storage

    sqla_sqlite = sql_storage.create_engine('asqlite')
    sqla_pg = sql_storage.create_engine('apg')
    sqla_mysql = sql_storage.create_engine('ams')



from typing import Set, List, Optional, Dict, Any

from .base import ModelProtocol
from .fields import ModelFieldsMixin
from ..serializers import Serializer, Deserializer
from ..serializers.pickle import PickleSerializer, PickleDeserializer


class SerializableModel(ModelFieldsMixin):
    class Meta(ModelFieldsMixin.Meta):
        serializer: Serializer
        deserializer: Deserializer

    def __init_subclass__(
            cls, serializer: Optional[Serializer] = None,
            deserializer: Optional[Deserializer] = None,
            **kwargs):
        super().__init_subclass__(**kwargs)
        cls.Meta.serializer = serializer or PickleSerializer()
        cls.Meta.deserializer = deserializer or PickleDeserializer()

    # noinspection PyProtectedMember
    @classmethod
    def get_serializable_fields(cls) -> Set[str]:
        key = cls._fields_cache_key()
        if key not in cls.Meta._fields_cache:
            cls._calculate_fields()
        return cls.Meta._fields_cache[key].ser

    def get_serializable_values(self, only_fields: List[str] = None) -> Optional[Dict[str, Any]]:
        serializable_fields = self.get_serializable_fields()
        only_fields = set(only_fields) if only_fields else set()

        return {
            name: getattr(self, name) for name in serializable_fields
            if hasattr(self, name) and (not only_fields or name in only_fields)
        }

    def serialize(self, only_fields: List[str] = None) -> bytes:
        return self.Meta.serializer.serialize(self.get_serializable_values(only_fields))

    @classmethod
    def deserialize(cls, data: bytes) -> Optional[ModelProtocol]:
        if deserialized := cls.Meta.deserializer.deserialize(data):
            return cls._construct(deserialized)
        return None



import contextlib
from dataclasses import dataclass, replace
from typing import Dict, List, Optional, Tuple

from aioredis import Redis as RedisConnection

from dropland.data.context import get_context
from .engine import RedisStorageBackend, RedisStorageEngine


@dataclass
class Session:
    engine: RedisStorageEngine
    connection: RedisConnection


class ConnectionContext:
    def __init__(self):
        self.sessions: Dict[str, Session] = dict()


class SessionManager:
    def __init__(self, engine_factory: RedisStorageBackend):
        self._engine_factory = engine_factory

    def get_session(self, name: str) -> Optional[Session]:
        ctx = get_context()
        if '_redis_session' not in ctx:
            ctx._redis_session = ConnectionContext()
        return ctx._redis_session.sessions.get(name)

    def get_or_create_session(self, name: str) -> Tuple[bool, Optional[Session]]:
        if session := self.get_session(name):
            return False, session

        if engine := self._engine_factory.get_engine(name):
            return True, Session(
                engine=engine,
                connection=engine.new_connection())

        return False, None

    @contextlib.asynccontextmanager
    async def session_context(self, name: str):
        created, session = self.get_or_create_session(name)

        if not created:
            yield session
            return

        async with session.connection as conn:
            session = replace(session, connection=conn)
            yield self._add_session(name, session)
            self._remove_session(name)

    @contextlib.asynccontextmanager
    async def init_engines(self, names: List[str] = None):
        engines = self._engine_factory.get_engines(names or [])
        async with contextlib.AsyncExitStack() as stack:
            for name, engine in engines.items():
                assert engine is self._engine_factory.get_engine(name)
                if engine.is_async:
                    await engine.async_start()
                    stack.push_async_callback(engine.async_stop)
                else:
                    engine.start()
                    stack.callback(engine.stop)
            yield

    def _add_session(self, name: str, data: Session) -> Session:
        ctx = get_context()
        if '_redis_session' not in ctx:
            ctx._redis_session = ConnectionContext()
        ctx._redis_session.sessions[name] = data
        return data

    def _remove_session(self, name: str):
        ctx = get_context()
        if '_redis_session' in ctx:
            ctx._redis_session.sessions.pop(name)



import asyncio
from dataclasses import dataclass
from typing import Dict, List, Mapping, Optional

import aio_pika
from aio_pika.pool import Pool

from dropland.log import logger, tr
from ..base import StorageBackend, StorageEngine


@dataclass
class EngineConfig:
    url: str
    virtualhost: str = '/'
    timeout_seconds: int = 5
    pool_max_connections: int = 4
    pool_max_channels_per_connection: int = 100


class RmqStorageBackend(StorageBackend):
    def __init__(self):
        self._engines: Dict[str, 'RmqStorageEngine'] = dict()

    @property
    def name(self) -> str:
        return 'rabbitmq'

    def create_engine(self, name: str, config: EngineConfig) -> Optional['RmqStorageEngine']:
        if engine := self._engines.get(name):
            return engine

        engine = RmqStorageEngine(self, name, config)
        self._engines[name] = engine
        logger.info(tr('dropland.blocks.rmq.engine.created').format(name=name))
        return engine

    def get_engine(self, name: str) -> Optional['RmqStorageEngine']:
        return self._engines.get(name)

    def get_engine_names(self) -> List[str]:
        return list(self._engines.keys())

    def get_engines(self, names: Optional[List[str]] = None) -> Mapping[str, 'RmqStorageEngine']:
        engines = dict()

        if not names:
            names = self.get_engine_names()

        for name in names:
            if engine := self.get_engine(name):
                engines[name] = engine

        return engines


class RmqStorageEngine(StorageEngine):
    def __init__(self, backend: RmqStorageBackend, name: str, config: EngineConfig):
        super().__init__(backend)
        self._name = name
        self._config = config
        self._lock = asyncio.Lock()
        self._conn_pool: Optional[Pool] = None
        self._chan_pool: Optional[Pool] = None
        self._counter = 0

    @property
    def name(self) -> str:
        return self._name

    @property
    def is_async(self):
        return True

    def new_connection(self):
        return self._chan_pool.acquire()

    def start(self):
        raise RuntimeError('Use async_start method')

    def stop(self):
        raise RuntimeError('Use async_stop method')

    async def async_start(self):
        async with self._lock:
            if 0 == self._counter:
                assert not self._conn_pool
                assert not self._chan_pool

                loop = asyncio.get_event_loop()
                self._conn_pool = Pool(self._get_connection, max_size=self._config.pool_max_connections, loop=loop)
                self._chan_pool = Pool(self._get_channel, max_size=self._config.pool_max_channels_per_connection, loop=loop)
                logger.info(tr('dropland.blocks.rmq.engine.started'))

            self._counter += 1

    async def async_stop(self):
        async with self._lock:
            if 1 == self._counter:
                assert self._conn_pool
                assert self._chan_pool

                await self._chan_pool.close()
                await self._conn_pool.close()

                self._conn_pool = self._chan_pool = None
                logger.info(tr('dropland.blocks.rmq.engine.stopped'))

            self._counter = max(self._counter - 1, 0)

    async def _get_connection(self) -> aio_pika.abc.AbstractRobustConnection:
        return await aio_pika.connect_robust(
            self._config.url, virtualhost=self._config.virtualhost,
            timeout=self._config.timeout_seconds
        )

    async def _get_channel(self) -> aio_pika.Channel:
        async with self._conn_pool.acquire() as connection:
            return await connection.channel()



import pickle
from typing import Optional, Any

from . import Serializer, Deserializer


class PickleSerializer(Serializer):
    def serialize(self, data: Optional[Any] = None) -> bytes:
        return pickle.dumps(data)


class PickleDeserializer(Deserializer):
    def deserialize(self, data: bytes) -> Optional[Any]:
        try:
            values = pickle.loads(data) if data else None
        except (pickle.UnpicklingError, ValueError, ModuleNotFoundError, MemoryError):
            return None

        return values
